#include <NZNet.h>
#include <NZGateway.h>
#include <base64.h>

NZNet::NZNet(uint8_t* rxBuffer, uint8_t* txBuffer, NZRouter* router, WiFiClient* wifiClient, HTTPClient* httpClient){
    _rxMessage = (NZNet::Message*)rxBuffer;
    _txMessage = (NZNet::Message*)txBuffer;
    _router = router;
    _wifiClient = wifiClient;
    _httpClient = httpClient;
}

NZNet::NZNet(NZGateway* gateway){
    _wifiClient = &(gateway->wifiClient);
    _httpClient = &(gateway->httpClient);

    _rxMessage = (NZNet::Message*)(gateway->lora_recv_buf);
    _txMessage = (NZNet::Message*)(gateway->lora_send_buf);
    _router = &(gateway->router);

    _telemetryTable = &(gateway->telemetryTable);
    _paramTable = &(gateway->paramTable);

}

NZNet::NZNet(NZDevice* device){
    _rxMessage = (NZNet::Message*)(device->lora_recv_buf);
    _txMessage = (NZNet::Message*)(device->lora_send_buf);
    _router = &(device->router);

    _telemetryTable = &(device->telemetryTable);
    _paramTable = &(device->paramTable);
}

//Записывает команду ack() и возвращает размер сообщения 
uint8_t NZNet::ack(){
    _txMessage->opCode = 0;
    return 1;
};

//Записывает команду set_param(INT8 index, type value) и возвращает размер сообщения
uint8_t NZNet::set_param(uint8_t index, CustomParamTable& paramTable){
    _txMessage->opCode = 1;
    uint8_t* pos = _txMessage->data;
    *(pos++) = index;
    pos += paramTable[index].bufferWriteValue(pos);
    return pos - (uint8_t*)_txMessage;
};

uint8_t NZNet::set_param(uint8_t index, void* mem, uint8_t size){
    _txMessage->opCode = 1;
    uint8_t* pos = _txMessage->data;
    *(pos++) = index;
    memcpy(mem, pos, size);
    pos += size;
    return pos - (uint8_t*)_txMessage;
}

//Записывает команду telemetry(INT8 index, type value) и возвращает размер сообщения
uint8_t NZNet::send_data(uint8_t index, CustomFieldTable& telemetryTable){
    _txMessage->opCode = 2;
    uint8_t* pos = _txMessage->data;
    *(pos++) = index;
    pos += telemetryTable[index].bufferWriteValue(pos);
    return pos - (uint8_t*)_txMessage;
};

//Записывает команду set_param_full(type*n value) и возвращает размер сообщения
uint8_t NZNet::set_param_full(CustomParamTable& paramTable){
    _txMessage->opCode = 3;
    uint8_t* pos = _txMessage->data;

    //Проверка возможности пересылки параметров
    if (paramTable.transmitedDataSize() > TRANSMITABLE_MESSAGE_LEN) return 0;
    pos += paramTable.bufferWriteData(pos);

    return pos - (uint8_t*)_txMessage;
};

//Записывает команду send_data_full(type*n value) и возвращает размер сообщения
uint8_t NZNet::send_data_full(CustomFieldTable& telemetryTable){
    _txMessage->opCode = 4;
    uint8_t* pos = _txMessage->data;

    //Проверка возможности пересылки параметров
    if (telemetryTable.transmitedDataSize() > TRANSMITABLE_MESSAGE_LEN) return 0;
    pos += telemetryTable.bufferWriteData(pos);

    return pos - (uint8_t*)_txMessage;
};

//Записывает команду register(c_str name, c_str model) и возвращает размер сообщения
uint8_t NZNet::server_register(char* name, char* model){
    _txMessage->opCode = 5;
    uint8_t* pos = _txMessage->data;

    strcpy((char*)pos, name);
    pos += strlen(name) + 1;

    strcpy((char*)pos, model);
    pos += strlen(model) + 1;

    return pos - (uint8_t*)_txMessage;

};

//Записывает команду disconnect() и возвращает размер сообщения
uint8_t NZNet::disconnect(){
    _txMessage->opCode = 6;
    return 1;
};



void NZNet::set_param_handler(){

    uint8_t index = (_rxMessage->data)[0];
    _paramTable->operator[](index).updateValue(_rxMessage->data+1);

    Serial.printf("set_param_handler()\n");

    //TODO отладить установку параметра

    uint8_t messageLen = ack();
    _router->sendtoGatewayWait((uint8_t*)_txMessage, messageLen);
};

void NZNet::send_data_handler(uint16_t address){
    base64 b64;
    char output_json[256];
    
    uint8_t index = _rxMessage->data[0];
    String parameter = b64.encode(_rxMessage->data+1, TRANSMITABLE_MESSAGE_LEN-1);

    _jsonDoc["address"] = address;
    _jsonDoc["index"] = index;
    _jsonDoc["value"] = parameter; 

    serializeJson(_jsonDoc, output_json);

    _httpClient->begin(*_wifiClient, "http://192.168.1.100/lora/telemetry");
    _httpClient->addHeader("Content-type", "application/json");
    Serial.printf("send_data_handler, code : %d\n", _httpClient->sendRequest("POST", output_json));
    Serial.printf("%s\n", output_json);
    _httpClient->end();

};

void NZNet::set_param_full_handler(){
    Serial.printf("set_param_full_handler()\n");
    uint8_t* pos = _rxMessage->data;
    for (uint8_t i = 0; i < _paramTable->size; i++) {
        uint8_t fieldSize = _paramTable->operator[](i).valueSize();
        _paramTable->operator[](i).updateValue(pos);
        pos += fieldSize;   
    }
};

void NZNet::send_data_full_handler(uint16_t address){
    base64 b64;
    char output_json[JSON_BUFFER_SIZE];

    String parameter = b64.encode(_rxMessage->data, TRANSMITABLE_MESSAGE_LEN-1);
    Serial.println(parameter);

    _jsonDoc["address"] = address;
    _jsonDoc["value"] = parameter;

    serializeJson(_jsonDoc, output_json);    

    _httpClient->begin(*_wifiClient, "http://192.168.1.100/lora/telemetry_full");
    _httpClient->addHeader("Content-type", "application/json");
    Serial.printf("send_data_full_handler, code : %d\n", _httpClient->sendRequest("POST", output_json));
    Serial.printf("%s\n", output_json);
    _httpClient->end();
};

void NZNet::server_register_handler(uint16_t address){
    char output_json[256];

    _jsonDoc["address"] = address;
    _jsonDoc["name"] = _rxMessage->data;
    _jsonDoc["model"] = _rxMessage->data + strlen((char*)_rxMessage->data) + 1;

    serializeJson(_jsonDoc, output_json);

    _httpClient->begin(*_wifiClient, "http://192.168.1.100/lora/connect_device");
    _httpClient->addHeader("Content-type", "application/json");

    uint8_t code =  _httpClient->sendRequest("POST", output_json);
    Serial.printf("register_handler, code : %d\n", _httpClient->sendRequest("POST", output_json));
    Serial.printf("%s\n", output_json);
    if (code == 200) {
        _router->sendtoDeviceWait((uint8_t*)_txMessage, ack(), address);
    }
    _httpClient->end();

};

void NZNet::disconnect_handler(uint16_t address){
    char output_json[256];
    _router->deleteRouteTo(address);
    //Если не шлюз
    if (_wifiClient == NULL) return;

    _jsonDoc["address"] = address;
    
    serializeJson(_jsonDoc, output_json);

    _httpClient->begin(*_wifiClient, "http://192.168.1.100/lora/disconnect");
    _httpClient->addHeader("Content-type", "application/json");
    Serial.printf("register_handler, code : %d\n", _httpClient->sendRequest("POST", output_json));
    Serial.printf("%s\n", output_json);
    _httpClient->end();

};


void NZNet::handle_rxMessage(uint16_t address){

    _jsonDoc.clear();


    switch (_rxMessage->opCode)
    {
    case 0: break;
    case 1: set_param_handler(); break;
    case 2: send_data_handler(address); break;
    case 3: set_param_full_handler();
    case 4: send_data_full_handler(address); break;
    case 5: server_register_handler(address); break;
    case 6: disconnect_handler(address); break;
    
    default:
        break;
    }
}