#include <NZRouter.h>
#define DEBUGPRINT

NZRouter::RoutedMessage NZRouter::_tmpMessage;

NZRouter::NZRouter(RHGenericDriver& driver, uint8_t localGateway, uint8_t thisAddress, uint16_t thisNetAddress) : RHReliableDatagram(driver, thisAddress){
    _thisNetworkAddress = thisNetAddress;
    _thisGateway = localGateway;
    clearRoutingTable();
}

bool NZRouter::init() {
    bool result  = RHReliableDatagram::init();
    return result;
}

void NZRouter::addRouteTo(uint8_t destNetAddr, uint8_t next_hop_LoRa, uint8_t state) {

    #ifdef DEBUGPRINT
        Serial.printf("Routing table updated! %d -> %d;\n", destNetAddr, next_hop_LoRa);
    #endif


    for (int i = 0; i < ROUTING_TABLE_SIZE; i++){
        if (_routes[i].destNetAddr == destNetAddr) {
            _routes[i].destNetAddr = destNetAddr;
            _routes[i].next_hop_LoRa = next_hop_LoRa;
            _routes[i].state = state;
            _routes[i].last_timestamp = millis() >> 16;
            _routingTableUsage++;
            return;
        }
    }

    if (_routingTableUsage == ROUTING_TABLE_SIZE) retireOldestRoute();

    for (int i = 0; i < ROUTING_TABLE_SIZE; i++){
        if (_routes[i].state == Unknown) {
            _routes[i].destNetAddr = destNetAddr;
            _routes[i].next_hop_LoRa = next_hop_LoRa;
            _routes[i].state = state;
            _routes[i].last_timestamp = millis() >> 16;
                _routingTableUsage++;
                return;
            }
        }


}

NZRouter::RoutingTableEntry* NZRouter::getRouteTo(uint16_t dest){
    for (int i = 0; i < ROUTING_TABLE_SIZE; i++)
        if (_routes[i].destNetAddr == dest && _routes[i].state != Unknown){
            return &_routes[i];
        }
    return NULL;
}

void NZRouter::printRoutingTable(){
    for (int i = 0; i < ROUTING_TABLE_SIZE; i++){
        Serial.printf("%d | Dest: %d Next_LoRa: %d State: %d Time: %d\n", i, _routes[i].destNetAddr, _routes[i].next_hop_LoRa, _routes[i].state, _routes[i].last_timestamp);
    }
}

bool NZRouter::deleteRouteTo(uint16_t addr){
    for (int i = 0; i < ROUTING_TABLE_SIZE; i++){
        if (_routes[i].destNetAddr == addr) {
            deleteRoute(i);
            return true;
        }
    }
    return false;
}

void NZRouter::retireOldestRoute(){
    int oldestIndex = 0;

    for (int i = 0; i < ROUTING_TABLE_SIZE; i++){
        if (_routes[i].last_timestamp < _routes[oldestIndex].last_timestamp)
            oldestIndex = i;
    }

    _routes[oldestIndex].state = Unknown;
    _routingTableUsage--;

    #ifdef DEBUGPRINT
        Serial.printf("Route %d -> %d removed;\n", _routes[oldestIndex].destNetAddr, _routes[oldestIndex].next_hop_LoRa);
    #endif

}

void NZRouter::clearRoutingTable(){
    for (int i = 0; i < ROUTING_TABLE_SIZE; i++) _routes[i].state = Unknown;
}

uint8_t NZRouter::sendtoGatewayWait(uint8_t* buf, uint8_t len){
    _tmpMessage.header.addrFlags = (_thisNetworkAddress << 2) + 1;
    memcpy(_tmpMessage.data, buf, len);

    return route(&_tmpMessage, sizeof(RoutedMessageHeader)+len);
}

uint8_t NZRouter::sendtoDeviceWait(uint8_t* buf, uint8_t len, uint16_t dest){
    _tmpMessage.header.addrFlags = dest << 2;
    memcpy(_tmpMessage.data, buf, len);

    return route(&_tmpMessage, sizeof(RoutedMessageHeader)+len);
}

uint8_t NZRouter::route(RoutedMessage* message, uint8_t messageLen){

    uint8_t next_hop = -1;
    uint8_t flags = message->header.addrFlags & 0b11;

    switch (flags){
        case 0:
        {  
            RoutingTableEntry* route = getRouteTo(message->header.addrFlags >> 2);
            if (!route) return RH_ROUTER_ERROR_NO_ROUTE;
            next_hop = route->next_hop_LoRa;
            route->last_timestamp = millis() >> 16;
            break;
        }
        case 1:
        {
            next_hop = _thisGateway;
            break;
        }
        case 2:
        {
            next_hop = _thisGateway;         
            break;
        }          
    }

    #ifdef DEBUGPRINT
        Serial.printf("Message routed, next hop is %d; %d %d\n", next_hop, flags, message->header.addrFlags >> 2);
        peekAtMessage(&_tmpMessage, messageLen);
    #endif

    if (!RHReliableDatagram::sendtoWait((uint8_t*)&_tmpMessage, messageLen, next_hop))
        return RH_ROUTER_ERROR_UNABLE_TO_DELIVER;

    #ifdef DEBUGPRINT
        Serial.printf("Message routed, successfully\n");
    #endif

    return RH_ROUTER_ERROR_NONE;

}

bool NZRouter::recvfromAck(uint8_t* buf, uint8_t* len, uint16_t* addrFlags, uint8_t* id){
    uint8_t tmpMessageLen = 255;
    uint8_t _from;
    uint8_t _to;
    uint8_t _id;
    uint8_t _flags;

    if (RHReliableDatagram::recvfromAck((uint8_t*)&_tmpMessage, &tmpMessageLen, &_from, &_to, &_id, &_flags)){

        //peekAtMessage(&_tmpMessage, tmpMessageLen);

        uint8_t flags = _tmpMessage.header.addrFlags & 0b11;
        uint16_t addr = _tmpMessage.header.addrFlags >> 2;

        if ((flags == 0 && addr == _thisNetworkAddress) || (flags == 1 && _thisNetworkAddress == 0) || (flags == 2)){

            if (flags == 0){
                
                addRouteTo(0, _from, Active);
                #ifdef DEBUGPRINT
                    Serial.printf("Got message for current node, source: GATEWAY, from: %d, len: %d\n", _from, tmpMessageLen);
                #endif

            }
            else {

                addRouteTo(addr, _from, Active);               
                #ifdef DEBUGPRINT
                    Serial.printf("Got message for current node, source: %d, from: %d, len: %d\n", addr, _from, tmpMessageLen);
                #endif 
            }

           

            //Если доставлено по адресу
            if (addrFlags) *addrFlags = _tmpMessage.header.addrFlags;

            uint8_t msgLen = tmpMessageLen - sizeof(_tmpMessage.header.addrFlags);
            if (*len > msgLen) *len = msgLen;
            memcpy(buf, _tmpMessage.data, *len);
            return true;
        }
        else {

            if (flags == 1 || flags == 2) {
                addRouteTo(addr, _from, Active);
            }

            #ifdef DEBUGPRINT
                if (flags == 1) Serial.printf("Got message for GATEWAY, rerouting...\n");
                else Serial.printf("Got message for %d, rerouting...\n", addr);
            #endif


            //Иначе переслать дальше
            route(&_tmpMessage, tmpMessageLen);
        }

    }
    return false;

}

bool NZRouter::recvFromAckTimeout(uint8_t* buf, uint8_t* len, uint16_t timeout, uint16_t* addrFlags, uint8_t* id){

    uint32_t starttime = millis();
    int32_t time_left;
    while ((time_left = timeout - (millis() - starttime)) > 0){
        if (waitAvailableTimeout(time_left)){
            if (recvfromAck(buf, len, addrFlags, id)) return true;
        }
        YIELD;
    }

    return false;

}

void NZRouter::peekAtMessage(RoutedMessage* message, uint8_t messageLen){
    printBufferedMessage(messageLen);
    //(void)message;
    //(void)messageLen;
}

void NZRouter::setThisNetworkAddress(uint16_t newAddress) {
    _thisNetworkAddress = newAddress;
}

void NZRouter::setThisLocalGateway(uint8_t newLocalGateway){
    _thisGateway = newLocalGateway;
}

void NZRouter::deleteRoute(uint8_t index){

    _routes[index].state = Unknown;
}

void NZRouter::printBufferedMessage(uint8_t tmpMessageLen){
    Serial.printf("Addr: %d, flags: %d, len: %d\n", _tmpMessage.header.addrFlags >> 2, _tmpMessage.header.addrFlags & 3, tmpMessageLen);
    for (int i = 0; i < tmpMessageLen; i++) Serial.printf("%02x ", ((uint8_t*)&_tmpMessage)[i]);
    Serial.println();
}

uint16_t NZRouter::thisNetworkAddress(){ return _thisNetworkAddress;}

uint16_t NZRouter::headerAddr(){
    return _tmpMessage.header.addrFlags >> 2;
}